// Copyright (c) 2022 by István Váradi

// This file is part of VSCPL, a simple cross-platform utility library

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

#pragma once

//------------------------------------------------------------------------------

#include "WSAInterface.h"

#include "BufferedStream.h"
#include "SocketMixin.h"
#include "SocketEventFactory.h"

//------------------------------------------------------------------------------

namespace hu { namespace varadiistvan { namespace scpl { namespace io {

//------------------------------------------------------------------------------

/**
 * A socket that works on a TCP connection.
 */
class TCPSocket : public SocketMixin<BufferedStream>
{
protected:
    /**
     * Construct the TCP socket.
     */
    TCPSocket(Waiter* waiter = 0, int protocol = 0,
              size_t readingCapacity = DEFAULT_CAPACITY,
              size_t writingCapacity = DEFAULT_CAPACITY);

    /**
     * Construct the socket with the given parameters.
     */
    TCPSocket(HANDLE s, Waiter* waiter = 0,
              size_t readingCapacity = DEFAULT_CAPACITY,
              size_t writingCapacity = DEFAULT_CAPACITY);

public:
    /**
     * Destroy the socket by closing the file descriptor
     */
    ~TCPSocket();

protected:
    /**
     * Connect the client socket to the server.
     */
    bool connect(const struct sockaddr_in& sin);

    friend class TCPAcceptor;
    friend class TCPConnector;
};

//------------------------------------------------------------------------------
// Inline definitions
//------------------------------------------------------------------------------

inline TCPSocket::TCPSocket(Waiter* waiter, int protocol,
                            size_t readingCapacity,
                            size_t writingCapacity) :
    TCPSocket(WSAInterface::get().socket(AF_INET, SOCK_STREAM, protocol),
              waiter, readingCapacity, writingCapacity)
{
}

//------------------------------------------------------------------------------

inline TCPSocket::TCPSocket(HANDLE s, Waiter* waiter,
                            size_t readingCapacity,
                            size_t writingCapacity) :
    SocketMixin<BufferedStream>(std::make_unique<SocketEventFactory>(),
                                waiter, s,
                                readingCapacity, writingCapacity)
{
}

//------------------------------------------------------------------------------

inline TCPSocket::~TCPSocket()
{
    if (!WSAInterface::get().isInvalidSocket(handle)) {
        WSAInterface::get().closesocket(handle);
    }
}


//------------------------------------------------------------------------------

inline bool TCPSocket::connect(const struct sockaddr_in& sin)
{
    auto a = WSAInterface::get().connect(handle,
                                         reinterpret_cast<const struct sockaddr*>(&sin),
                                         sizeof(sin));
    if (a==SOCKET_ERROR) {
        auto lastError = GetLastError();
        if (lastError!=WSAEWOULDBLOCK && lastError!=WSAEINPROGRESS) {
            setErrorCode(lastError);
        }
    }

    return a!=SOCKET_ERROR;
}

//------------------------------------------------------------------------------

} /* namespace hu::varadiistvan::scpl::io */ } /* namespace hu::varadiistvan::scpl */ } /* namespace hu::varadiistvan */ } /* namespace hu */

//------------------------------------------------------------------------------

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
